/****************************************************************************
 *
 *   File: analyze.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Thu Jul 17 18:53:03 2003
 *
 *   Function: This tool collects statistics information for a given
 *             priority function.  It essentially compiles the given
 *             applications with the priority function, specifying to
 *             the finch library to annotate the priority function with
 *             profile information-- or data collected when the expression
 *             is evaluated (e.g., range information, stddev, etc.).
 *
 *             This program collects and merges the annotations.  It 
 *             then prints statistical information about the expression
 *             to a dot-readable format.
 *
 ***************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <math.h>
#include <bits/nan.h>
#include "benchmark_set.h"
#include "cmdparse.h"
#include "genetic_expr.h"
#include "exec_utils.h"


void
dispatch (const string& workdir, 
	  benchmark_info *bench, 
	  genetic_expr *expr)
/////////////////////////////////////////////////////////////
//  
//  Actually compiles and runs the benchmark with the given
//  expression.  All fitness bookkeeping is stored in the
//  genetic_expr and the benchmark_info.
//
/////////////////////////////////////////////////////////////
{
  double fitness;
  char *curdir;

  // Check to see if the benchmark hasn't
  // already been memoized.
  cout << "Bench=" << bench->bid() << ", ";

  string eval_dir = workdir + string("/evaluate/");
  string expr_fname = eval_dir + string ("pfun.l");
  string fit_fname = eval_dir + string ("fitness");

  // Save away the current directory and switch to 
  // the compile directory.
  curdir = getcwd (NULL, 0);
  chdir (eval_dir.c_str());

  // Execute the setup commands first.
  execute_commands (bench->get_setup_commands());

  // Write the expression to $FINCH_HOME/evaluate/pfun.l
  ofstream expr_file (expr_fname.c_str());
  assert (expr_file);
  expr->print (expr_file);
  expr_file.close ();
  
  // Execute all the commands needed to profile, compile
  // run, and check the benchmark.  It is the user's 
  // responsibility to make sure that the fitness file
  // is written.
  execute_commands (bench->get_profile_commands());
  execute_commands (bench->get_compile_commands());
  execute_commands (bench->get_link_commands());
  execute_commands (bench->get_run_commands());
  
  ifstream fitfile (fit_fname.c_str());
  fitness = NAN;
  if (!fitfile.fail()) {
    fitfile >> fitness;
    fitfile.close ();
  }
  
  execute_commands (bench->get_check_commands());
  
  cout << "Fitness=" << fitness << ">\n";
  
  // Assign the fitness to the expression, thereby
  // memoizing it.
  if ((fitness == 0.0) || isnan(fitness))
    expr->set_fitness_element (bench->bid(), NAN);
  else 
    expr->set_fitness_element (bench->bid(), fitness);
  
  finch_clean_dir (eval_dir, NULL);
  
  // Switch back to the harness directory.
  chdir (curdir);
  delete[] curdir;
}


typedef void (*map_pdata_fun)(const char*, genetic_expr*);

void collect_data  (char*, genetic_expr*, benchmark_set*);
int  select_pfiles (const struct dirent*);
void process_notes_files (const string&, genetic_expr*, map_pdata_fun);
void remove_file (const char*, genetic_expr*);
void merge_data (const char*, genetic_expr*);
void compute_stats (genetic_expr *e, char *dotfname);

genetic_expr_bnf* g_bnf;

int
main (int32 argc, char **argv)
/*************************************************************
 *
 *  Parses the command line, and runs the analyzer.
 *
 ************************************************************/
{
  bool   help;
  char*  expr_fname;
  char*  benches_fname;
  char*  dot_fname;
  char*  finchhome;
  uint   srate;

  genetic_expr*     expr;
  benchmark_set*    benches;

  help = false;

  static cmd_line_option options[] = {
    {CLO_NOARG, "-help",    "", &help},
    {CLO_STRING,"-benches", "", &benches_fname},
    {CLO_STRING,"-dotfname", "stats.dot", &dot_fname},
    {CLO_STRING,"-expr",    "", &expr_fname},
    {CLO_INT,   "-srate", "10", &srate},
    {CLO_NOARG, "-verbose", "", &g_verbose}
  };

  for (int32 i = 0; i < argc; i++)
    cout << argv[i] << " ";
  cout << endl;

  parse_cmd_line (argc, argv, options, 
		  sizeof(options)/sizeof(cmd_line_option));

  if (!expr_fname || !benches_fname || help) {
    cout << "\nUsage: " << argv[0];
    cout << " [options] -expr <expression_file> -benches <benchmarks_file>\n\n";
    if (help) {
      cout << "\n\toptions: \n";
      cout << "\t -dotfname   the dot output file name.\n";
      cout << "\t -help       print help information.\n";
      cout << "\t -srate      the sample rate [1..100].\n";
      cout << "\t -verbose    puts analyzer into verbose mode.\n\n";
    }
    exit (1);
  }

  // Make sure that the FINCH_HOME environment variable is set.
  finchhome = getenv ("FINCH_HOME");
  finch_claim (finchhome != NULL, 
	       "Please set the FINCH_HOME environment variable.");

  cout << "\nAnalysis tool.\nFinch Version 5-05-2004\n\n";

  finch_claim (expr_fname[0] != '\0', "Invalid expression file.");
  finch_claim (benches_fname[0] != '\0', "Invalid benchmark file.");
  finch_claim (srate > 0, "Sample rate must be greater than 0.");
  finch_claim (srate <= 100, "Sample rate must be less than 100.");

  // The finch library will look to see if the file 
  // $FINCH_HOME/srate exists.  If it does, it will
  // sample according to the rate specified by the 
  // file.
  string srate_fname = string(finchhome) + string ("/srate");
  ofstream srate_file (srate_fname.c_str());
  srate_file << (double) ((double)srate / 100.0) << endl;
  srate_file.close ();
  
  // We need a bnf with which to parse the expression, but the
  // number of parameters doesn't have to be accurate.
  g_bnf = new genetic_expr_bnf_default (0, 0);

  // Open the expression.
  ifstream expr_file (expr_fname);
  expr = genetic_expr::parse_expr (g_bnf, expr_file);

  // Create the benchmarks class.
  benches = new benchmark_set (benches_fname);

  // Clear the notes in the notes directory if there is one.
  // Make one if necessary.
  string notes_dir = string(finchhome) + string("/notes/");
  mkdir (notes_dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
  process_notes_files (notes_dir, expr, remove_file);

  // Make an evaluate directory in case there isn't one 
  // already there.
  string eval_dir = string(finchhome) + string("/evaluate");
  mkdir (eval_dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);

  // Run each of the benchmarks and collect the data.
  collect_data (finchhome, expr, benches);

  // Merge the data together.
  process_notes_files (notes_dir, expr, merge_data);

  // Computes stats.
  compute_stats (expr, dot_fname);

  // Print out the expression with all the merged data.
  ofstream annotated_expr("annotated_expr.l");
  expr->print (annotated_expr, true);

  delete g_bnf;
  delete expr;
}


void
process_notes_files (const string& notes_dir, 
		     genetic_expr *expr,
		     map_pdata_fun fun)
  /****************************************************
   *  This function finds all files that start with
   *  _note, and processes them with the map function
   *  that is passed in.
   ***************************************************/
{
  int numpfiles;
  struct dirent** pfilelist;

  numpfiles = scandir (notes_dir.c_str(), 
		       &pfilelist, 
		       select_pfiles, 0);

  for (int i = 0; i < numpfiles; i++)
  {
    string delfile = notes_dir + string(pfilelist[i]->d_name);
    fun (delfile.c_str(), expr);
    free (pfilelist[i]);
  }
  free (pfilelist);
}


void
remove_file (const char *expr_fname, genetic_expr *e)
  /****************************************************
   *  This map_pdata_fun simply unlinks the file.
   ***************************************************/
{
  unlink (expr_fname);
}


void
merge_data (const char *expr_fname, genetic_expr *e)
  /****************************************************
   *  This function reads in an annotated expression
   *  and merges it with expression e.
   ***************************************************/
{
  genetic_expr *tmp;

  tmp = genetic_expr::parse_expr (g_bnf, expr_fname);
  assert (tmp && e);
  e->merge (tmp,true);
}


void
compute_stats (genetic_expr *e, char *dotfname)
  /****************************************************
   *  This function calls the stat_print function in
   *  genetic_expr.h.  This function prints out 
   *  a file that can be processed by 'dot'.  The
   *  final output is a graphical depiction of the
   *  AST, annotated with statistical information 
   *  about each expression.
   ***************************************************/
{
  ofstream dotfile(dotfname);
  e->stat_print (dotfile);
  dotfile.flush();
}	    


void
collect_data (char *finchhome,
	      genetic_expr *expr, 
	      benchmark_set *bset)
  /****************************************************
   *  This runs the compiler to collect statistics
   *  information.
   ***************************************************/

{
  benchmark_info *bench;

  cout << "Collecting data.\n";
  for (uint32 bn = 0; bn < bset->num_benches(); bn++)
  {
    cout << "<";
    bench = bset->bench (bn);
    dispatch (finchhome, bench, expr);
  }
}


int
select_pfiles (const struct dirent* entry)
  /****************************************************
   *  Returns true if the file is an annotes file.
   ***************************************************/
{
  int sel = !strncmp(entry->d_name, "_note", 5);
  return sel;
}

