/*************************************************************************
 *
 *   File: compile_harness.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Wed Aug 21 19:58:23 2002
 *
 *   Function: base class for the various harnesses we use.
 *
 ************************************************************************/

#include "compile_harness.h"
#include "exec_utils.h"
#include <unistd.h>
#include <dirent.h>
#include <algorithm>
#include <stdio.h>
#include <signal.h>
#include <sys/stat.h>

compile_harness::compile_harness (uint32  population_size,
				  uint32  generations,
				  double  mutate_rate,
				  double  mortality,
				  char*   initial_pop_fname,
				  char*   save_pop_fname,
				  char*   finch_home,
				  uint32  tournament_size,
				  uint32  dss_set_size,
				  uint32  maximum_genome_size,
				  uint32  max_height,
				  uint32  elitist,
				  uint32  sunset,
				  bool    quantize_results,
				  bool    classifier,
				  benchmark_set* benches) :

  m_workdir(finch_home),
  m_gens(generations),
  m_mutate_rate(mutate_rate),
  m_mortality(mortality),
  m_save_pop_fname(save_pop_fname),
  m_state_file(NULL),
  m_tournament_size(tournament_size),
  m_dss_set_size(dss_set_size),
  m_max_genome_size(maximum_genome_size),
  m_elitist(elitist),
  m_sunset(sunset),
  m_benches(benches),
  m_quantize_results(quantize_results),
  m_classifier(classifier)
{
  // Register a signal handler to quit finch with C-c.
  signal(SIGINT, finch_sigint);

  finch_claim (benches, "Missing benchmark set.");

  cout << "Population size    : " << population_size << endl;
  cout << "Generations        : " << generations << endl;
  cout << "Mutation rate      : " << mutate_rate << endl;
  cout << "Mortality          : " << mortality << endl;
  cout << "Initial population : " << initial_pop_fname << endl;
  cout << "Last population    : " << save_pop_fname << endl;
  cout << "Tournament size    : " << tournament_size << endl;
  cout << "DSS set size       : " << dss_set_size << endl;
  cout << "Max genome size    : " << maximum_genome_size << endl;
  cout << "Maximum height     : " << max_height << endl;
  cout << "Elitism            : " << elitist << endl;
  cout << "Quantize results   : " << quantize_results << endl;
  cout << "Classifier         : " << classifier << endl;
  cout << "$FINCH_HOME        : " << finch_home << endl;

  struct stat eval_stat;
  stat(finch_home, &eval_stat);
  finch_claim (S_ISDIR(eval_stat.st_mode), 
	       "Make sure that the directory $FINCH_HOME exists");

  // Append /evaluate/ to the user-given evaluate directory, because
  // some of the harnesses clear this directory, and if the user
  // were to inadvertently supply a directory he cared about,
  // it would delete his precious files.
  m_run_dir = string(finch_home);
  m_eval_dir = m_run_dir + string("/evaluate/");

  // To make this as easy for the user as possible, a configuration
  // file is automatically created.  The configuration file contains
  // information such as the number of double arguments and boolean
  // arguments that the priority function should consider.  This
  // harness creates this file in a few easy steps.
  // 
  // 1. Delete the configuration file in $FINCH_EVAL_DIR/config
  // 2. Invoke the compiler
  // 3. The finch library call, evaluate, will create the 
  //    $FINCH_EVAL_DIR/config file, then immediately exit.
  // 4. At this point, finch has the vital information that it
  //    needs to continue.
  create_configuration ();

  // Next, we initialize the population.  If there is a population
  // file, initialize the population with the file.  Otherwise,
  // create an initially random population using the ramped
  // half and half method.
  create_initial_population (population_size, 
			     max_height,
			     initial_pop_fname);

  // If the quantize results option is specified, keep a set
  // of fitnesses.
  if (m_quantize_results) m_quanta = new set<double>;
}


compile_harness::~compile_harness ()
/////////////////////////////////////////////////////////////
//
//  Deletes dynamically allocated memory.
//
/////////////////////////////////////////////////////////////
{
  if (m_quantize_results) delete m_quanta;

  for (uint32 i = 0; i < popsize(); i++)
    delete m_population[i];

  delete m_bnf;
}



genetic_expr*
compile_harness::search_for_best ()
/////////////////////////////////////////////////////////////
//
//  This is the function that iteratively searches for the
//  best priority function.
//
/////////////////////////////////////////////////////////////
{
  char savepop_fname[80];
  char state_fname[80];
  uint32 gen;

  // Clear the work directories.
  clear_evaluate_dir ();

  m_memoized = 0;
  for (gen = 1; gen <= m_gens; gen++) 
  {
    cout << "\nGeneration: " << gen;
    cout << ", " << m_population[0]->get_fitness() << endl;

    // Save all intermediate state.
    sprintf (state_fname, "_state_%d", gen);
    m_state_file = new ofstream(state_fname);
    assert (m_state_file);

    print_population (*m_state_file, (int32)popsize(), true);

    // Reset the fitnesses.
    prepare_for_generation ();

    // Farms out the jobs.
    farm_out_jobs ();

    // Sort the expression list in descending fitness.
    sort (m_population.begin(), m_population.end(), expr_lt());

    // Print the population.
    print_population (cout, 5);

    // Save away the best expression.
    ofstream best ("best.l");
    finch_claim (best.is_open(), "Couldn't write expression");
    m_population[0]->print (best);
    best.close ();

    m_population[0]->mark_as_best (m_sunset);
    cout << "Marked the best expression.\n";

    // Quantize the population if desired.
    if (m_quantize_results) {
      cout << "Quantizing population...\n";
      quantize_population ();
    }

    // Breed and randomly mutate expressions.
    crossover_and_mutate ();

    // Close the state file.
    m_state_file->close ();
    delete m_state_file;
  }

  // If specified, save away the entire population.
  sprintf (savepop_fname, "%s-%i", m_save_pop_fname, gen);
  ofstream savepop_file(savepop_fname);
  print_population (savepop_file, (int32)popsize());
  savepop_file.close();

  cout << endl;
  return m_population[0];
}


void
compile_harness::create_configuration ()
/////////////////////////////////////////////////////////////
//
// This function performs the following:
// 
// 1. Delete the configuration file in $FINCHHOME/config
// 2. Invoke the compiler
// 3. The finch library call, evaluate, will create the 
//    $FINCHHOME/config file, then immediately exit.
// 4. At this point, finch has the vital information that it
//    needs to continue.
//
/////////////////////////////////////////////////////////////
{
  benchmark_info *bench;
  uint32 num_dbls;
  uint32 num_bools;
  char *curdir;

  finch_claim ((m_benches->num_benches() > 0), "No training benchmarks.");
  bench = m_benches->bench (0);

  cout << "\nCreating configuration file.\nThis may take a minute...\n";

  // Create the config filename.
  string filename = m_workdir + string("/config");
  unlink (filename.c_str());

  // Save away the current directory and switch to 
  // the compile directory.
  curdir = getcwd (NULL, 0);

  // Compile in the eval dir.
  string eval_dir = m_eval_dir;
  mkdir (eval_dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
  chdir (eval_dir.c_str());
  
  // Execute all the commands needed to profile, compile
  // run, and check the benchmark.  It is the user's 
  // responsibility to make sure that the fitness file
  // is written.
  execute_commands (bench->get_setup_commands());
  execute_commands (bench->get_profile_commands());
  execute_commands (bench->get_compile_commands(), true);

  ifstream exists(filename.c_str());
  finch_claim (exists, "Could not create configuration file.");

  finch_claim ((exists >> num_dbls), "Bad configuration.");
  finch_claim ((exists >> num_bools), "Bad configuration.");
  exists.close ();

  cout << "Optimizing a priority function with:\n";
  cout << "\t" << num_dbls << " Doubles\n\t" << num_bools
       << " Booleans\n";
  
  // Create the bnf using these parameters.
  m_bnf = new genetic_expr_bnf_default (num_dbls, num_bools);

  // Change back to the directory we were in.
  chdir (curdir);
  delete[] curdir;
}

void
compile_harness::create_initial_population (uint32 population_size,
					   uint32 max_height,
					   char *initial_pop_fname)
/////////////////////////////////////////////////////////////
//
//  Initialize the population.
//
/////////////////////////////////////////////////////////////
{
  genetic_expr *tmp;
  uint32  expr_num;
  uint32  bid;
  uint32  interval;
  double  fitness;
  char    atchar;

  cout << endl;

  // Initialize the population.
  if (initial_pop_fname[0] != '\0')
  {
    ifstream initial_population (initial_pop_fname);
    finch_claim (initial_population.is_open(), 
		 "Can't open population.");

    // After every generation, the harness writes the population
    // to disk.  The expressions in the population are annotated
    // with the memoized fitnesses.  In addition partial state
    // during the generation is written to disk.  Such state is 
    // in the form:
    // @ <expression number> <benchmark id> <fitness>

    // An @ will therefore stop parsing of the expressions.
    while ((tmp = 
	    genetic_expr::parse_expr(m_bnf, initial_population)))
      m_population.push_back (tmp);

    finch_claim ((population_size >= popsize()),
	   "Initial population contains too many expressions.");

    // Put back the @ that ended the parsing of expressions, 
    // and read in the partial state.
    initial_population.unget ();
    initial_population.setf (ios::skipws);
    while (initial_population >> atchar && atchar == '@')
    {
      initial_population >> expr_num;
      initial_population >> bid;
      initial_population >> fitness;
      finch_claim ((expr_num < popsize()), 
		   "Invalid expression number.");
      finch_claim ((bid < m_benches->num_benches()), 
		   "Invalid benchmark number.");
      if (fitness > 0.0)
	m_population[expr_num]->set_fitness_element (bid, fitness);
      else
	cerr << "Warning: bad fitness discovered, discarding...\n";
    }

    initial_population.close ();
  }

  // If there are fewer expressions in the population file, 
  // populate the rest of the population with random expressions,
  // using the ramped-half-and-half method.
  if (max_height < 1){ 
    max_height = 1;
    interval = population_size;
  }
  else 
    interval = (population_size - popsize()) / (max_height - 1);
  
  cout << "\nCreating random expressions...\n\n";
  for (uint32 i = popsize(); i < population_size; i++) {
    if ((i % interval) == 0) max_height--;
    m_population.push_back (genetic_expr::random(m_bnf, 
						 max_height, 
						 NULL, 
						 m_classifier ? BOOLEAN : DOUBLE));
  }
}



void
compile_harness::crossover_and_mutate ()
/////////////////////////////////////////////////////////////
//
//  Performs crossover and mutation on the population.
//  This function is virtual so you can define your
//  own crossover and mutation schemes if you want.
//
/////////////////////////////////////////////////////////////

{
  genetic_expr *kill1;
  genetic_expr *kill2;
  genetic_expr *new1;
  genetic_expr *new2;
  uint32 kill1_ind;
  uint32 kill2_ind;
  uint32 parent1;
  uint32 parent2;
  uint32 rand;
  uint32 kill;

  uint32 kill_size = (uint32) (m_mortality * (double)popsize());

  // Since the array is sorted, just take the smallest of 
  // the randomly generated numbers.
  for (kill = 1; kill <= kill_size; kill += 2) {
    parent1 = popsize ();
    parent2 = popsize ();
    for (uint32 i = 0; i < m_tournament_size; i++) {
      rand = random() % popsize();
      assert (rand < popsize());
      if (rand <= parent1) parent1 = rand;
      rand = random() % popsize();
      assert (rand < popsize());
      if (rand <= parent2) parent2 = rand;
    }
    
    assert ((parent1 != popsize()) && (parent2 != popsize()));

    // Choose two different expressions to kill, but
    // spare all expressions between 0 and m_elitist.
    // Also spare any individuals that were the best
    // within the last SUNSET generations.
    do {
      kill1_ind = random() % (popsize()-m_elitist);
      kill1_ind += m_elitist;
      assert (kill1_ind < popsize());
    } while (m_population[kill1_ind]->was_best());

    do {
      kill2_ind = random() % (popsize()-m_elitist);
      kill2_ind += m_elitist;
      assert (kill1_ind < popsize());
    } while ((kill2_ind == kill1_ind) ||
	     (m_population[kill2_ind]->was_best()));

    // Crossover the two expressions and create two new ones.
    m_population[parent1]->cross (m_population[parent2], new1, new2);

    // Make sure the new expression does not exceed the
    // maximum genome size. Discard those that do.
    if (new1->size() < m_max_genome_size)
    {
      // Mutate the expression according to the mutation 
      // rate.
      if (drand48() < m_mutate_rate)
	new1->mutate ();
      kill1 = m_population[kill1_ind];
      m_population[kill1_ind] = new1;
      delete kill1;      
    }
    else
    {
      delete new1;
    }

    // Make sure the new expression does not exceed the
    // maximum genome size.  Discard those that do.
    if (new2->size() < m_max_genome_size)
    {
      // Mutate the expression according to the mutation 
      // rate.
      if (drand48() < m_mutate_rate)
	new2->mutate ();
      kill2 = m_population[kill2_ind];
      m_population[kill2_ind] = new2;
      delete kill2;
    }
    else
    {
      delete new2;
    }
  }

  // Since we only generated new expressions from this
  // point to the end of the population, we know that
  // the rest of the population has been memoized.
  m_memoized = popsize() - kill;
}


void
compile_harness::print_population (ostream& os, 
				   uint32 top,
				   bool state) const
/////////////////////////////////////////////////////////////////
//
//  This function computes stats on the entire population, but
//  only prints out the top creatures.
//
/////////////////////////////////////////////////////////////////
{
  uint32 lifetime_sum = 0;
  uint32 lifetime;
  double fit_sum = 0.0;
  double sq_sum = 0.0;
  double fit_mean;
  double fit;
  double diff;
  double var;

  os << endl;
  for (uint32 i = 0; i < popsize(); i++) 
  {
    fit = m_population[i]->get_fitness ();
    fit_sum += isnan(fit) ? 0.0 : fit;
    lifetime = m_population[i]->lifetime ();
    lifetime_sum += lifetime;

    if (i < top) 
    {
      os << "// " << i << " : [" << fit << ",";
      os << lifetime << "]\n";
      if (m_population[i]->was_best())
	os << "// THIS WAS A WINNER.\n";
      m_population[i]->print (os, state);
      os << endl;
    }
  }

  // Print the average fitness and lifetime.
  fit_mean = fit_sum / (double) popsize();
  os << "\n // Mean fitness = " << fit_mean;
  os << "\n // Mean lifetime = " << 
    ((double)lifetime_sum / (double)popsize());

  // Print the median fitness.  If the population size
  // is odd, median is N/2, else median is arithmetic
  // mean of N/2 and N/2+1.
  fit = m_population[popsize()/2]->get_fitness ();
  if (!(popsize() & 0x1))
  {
    fit += m_population[(popsize()/2)+1]->get_fitness ();
    fit = fit / 2;
  }

  os << "\n // Median fitness = " << fit;

  // Compute the standard deviation of the fitnesses.
  for (uint32 i = 0; i < popsize(); i++)
  {
    fit = m_population[i]->get_fitness ();
    fit = isnan(fit) ? 0.0 : fit;
    diff = fit - fit_mean;
    sq_sum += diff * diff;
  }

  var = sq_sum / (double)(popsize() - 1);
  os << "\n // Stdev = " << sqrt(var);

  if (m_quantize_results)
    os << "\n // Quanta = " << m_quanta->size();
  os << "\n";
  os.flush ();  
}


void
compile_harness::prepare_for_generation ()
/////////////////////////////////////////////////////////////
//
//  This function increments some counters and resets the
//  fitness of new expressions.
//
/////////////////////////////////////////////////////////////
{
  for (uint32 j = 0; j < popsize(); j++) {
    m_population[j]->inc_lifetime ();
    m_population[j]->set_fitness (0);
  }
}


void
compile_harness::quantize_population ()
/////////////////////////////////////////////////////////////
//
//  Records the number of different fitnesses that this
//  harness has seen.
//
//  Invariant: m_quantize_results must be true.
//
/////////////////////////////////////////////////////////////
{
  for (uint32 i = 0; i < popsize(); i++) {
    m_quanta->insert (m_population[i]->get_fitness());
  }
}


uint32
compile_harness::dss_size () const
/////////////////////////////////////////////////////////////
//
//  Returns the DSS set size to use.
//
/////////////////////////////////////////////////////////////
{
  uint32 setsize = m_dss_set_size;

  // Increase m_dss_set_size by -1, 0, or 1.
  if (setsize > 2) setsize += ((random() % 3) - 1);
  return (min(m_benches->num_benches(), setsize));
}


void
compile_harness::compute_overall_fitness (uint32 num_benches)
/////////////////////////////////////////////////////////////
//
//  Computes the overall fitness of each expression in 
//  the population.  Each expression is normalized by the
//  fitness of the 0th expression in the population.
//
/////////////////////////////////////////////////////////////
{
  uint32 bn;
  uint32 en;
  double fit_mean;
  double base_fitness;
  double fitness;
  double ssd;
  double diff;
  double stddev;
  double var;
  double num_exprs;
  benchmark_info *bench;
  genetic_expr *expr;

  // Finally, compute the overall fitness of each expression in 
  // the population.  This fitness is the percentage of execution
  // time of the 0th expression.
  for (bn = 0; bn < num_benches; bn++)
  {
    bench = benches()->bench (bn);

    fit_mean = 0.0;
    num_exprs = 1.0;
    base_fitness = population()[0]->get_fitness_element (bench->bid());
    assert(!isnan(base_fitness));
    for (en = 0; en < popsize(); en++) 
    {
      expr = population()[en];
      fitness = expr->get_fitness_element(bench->bid()) / base_fitness;
      if (!isnan(fitness))
      {
	num_exprs += 1.0;
	fit_mean += fitness;
      }
      expr->set_fitness (fitness + expr->get_fitness());
    }
    fit_mean = fit_mean / num_exprs;

    // Compute the standard deviation of fitnesses for
    // each benchmark.  If standard dev is low, then we
    // say that the benchmark is not optimizable.
    ssd = 0.0;
    for (en = 0; en < popsize(); en++)
    {
      expr = population()[en];
      fitness = expr->get_fitness_element (bench->bid());
      if (!isnan(fitness))
      {
	diff = fit_mean - (fitness / base_fitness);
	ssd += diff * diff;
      }
    }
    var = ssd / (double)(num_exprs - 1.0);
    stddev = (var <= 0) ? 0.0 : sqrt(var);

    bench->set_optimizability (stddev);
    bench->reset_recency ();
  }

  // Increment the recencies of all the benchmarks.
  benches()->increment_recencies ();
}



void
compile_harness::write_state (benchmark_info* bench,
			      genetic_expr *expr,
			      uint32 expr_num)
/////////////////////////////////////////////////////////////
//
//  Writes the fitness of the expression to the state
//  file.
//
/////////////////////////////////////////////////////////////
{
  (*m_state_file) << "@ " << expr_num << " "
		  << bench->bid() << " " 
		  << expr->get_fitness_element(bench->bid())
		  << endl;
}

