/*************************************************************************
 *
 *   File: pbs_harness.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Wed Aug 21 19:58:23 2002
 *
 *   Function: This harness uses a job queuing system to distribute
 *             jobs.  It uses the file system to collect fitnesses,
 *             and therefore relies on a networked file system (that
 *             ensures consistency).  It also relies on PBS, the 
 *             job queuing system we use.
 *
 ************************************************************************/

#include "pbs_harness.h"
#include "exec_utils.h"
#include "num_string.h"
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

extern "C" {
#include <pbs_error.h>
#include <pbs_ifl.h>
}

pbs_harness::pbs_harness (uint32  population_size,
			  uint32  generations,
			  double  mutate_rate,
			  double  mortality,
			  char*   initial_pop_fname,
			  char*   save_pop_fname,
			  char*   finch_home,
			  uint32  tournament_size,
			  uint32  dss_set_size,
			  uint32  maximum_genome_size,
			  uint32  maximum_height,
			  uint32  elitist,
			  uint32  sunset,
			  bool    quantize_results,
			  bool    classifier,
			  benchmark_set* benches) :
  compile_harness (population_size,
		   generations,
		   mutate_rate,
		   mortality,
		   initial_pop_fname,
		   save_pop_fname,
		   finch_home,
		   tournament_size,
		   dss_set_size,
		   maximum_genome_size,
		   maximum_height,
		   elitist,
		   sunset,
		   quantize_results,
		   classifier,
		   benches)
{
  create_attr_list ();
}


pbs_harness::~pbs_harness ()
/////////////////////////////////////////////////////////////
//
//  Deletes dynamically allocated memory.
//
/////////////////////////////////////////////////////////////
{
  struct attropl *first = m_common_attrs;
  while (first)
  {
    first = m_common_attrs->next;
    free (m_common_attrs);
    m_common_attrs = first;
  }
}


void
pbs_harness::clear_evaluate_dir ()
/////////////////////////////////////////////////////////////
//
// Clears old work done in the work directories.
// 
/////////////////////////////////////////////////////////////
{
  cout << "Clearining " << eval_dir() << endl;
  finch_clean_dir (eval_dir(), NULL);
}


void
pbs_harness::farm_out_jobs ()
/////////////////////////////////////////////////////////////
//
//  This procedure runs and collects fitness numbers for
//  all the expressions that haven't aleady been memoized.
//  This procedure is also responsible for determining 
//  and marking memoized expressions.
//
/////////////////////////////////////////////////////////////
{
  static char *server = "cagfarm-01";
  benchmark_info *bench;
  genetic_expr *expr;
  uint32 num_benches;
  uint32 bn;
  uint32 en;
  int32 ret;

  cout << endl;

  do
  {
    cerr << "Trying to connect to PBS server " << server << "...\n";
    m_pbs_stream = pbs_connect (server);
  } while (m_pbs_stream < 0);
  cerr << "Connected...\n";

  // Find the subset of benchmarks.
  benches()->sort ();

  // Summarize the DSS state in the transcript.
  benches()->summarize (cout);

  ofstream dssstate("_dssstate");
  if (dssstate) benches()->print (dssstate);
  dssstate.close ();

  // Figure out the setsize to use.
  num_benches = dss_size ();

  cout << "Using a set size of " << num_benches 
       << endl << endl;

  clear_evaluate_dir ();

  // Submit a whole generation worth of jobs.
  cerr << "\nSubmitting jobs to queue...\n\n";
  for (bn = 0; bn < num_benches; bn++)
  {
    bench = benches()->bench (bn);
    cerr << "Bench " << bench->bid() << ":\n";
    for (en = 0; en < popsize(); en++)
    {
      expr = population()[en];
      if (!(expr->memoized(bench->bid())))
	create_pbs_script (bench, expr, en);
    }
  }

  // Gather all of the results, cleaning up behind.
  cerr << "\nGathering results...\n\n";
  for (bn = 0; bn < num_benches; bn++)
  {
    bench = benches()->bench (bn);
    for (en = 0; en < popsize(); en++)
    {
      expr = population()[en];
      cout << bench->bid() << "\t" << en << "\t";
      if (!(expr->memoized(bench->bid()))) 
      {
	if (gather_results (bench, expr, en))
	{
	  cerr << "Warning: broken expression.\n";
	}
	else
	{
	  write_state (bench, expr, en);
	}
      }
      else
      {
	cout << expr->get_fitness_element(bench->bid()) 
	     << "(" << expr->lifetime() << ")\n";
      }
    }
  }

  compute_overall_fitness (num_benches);

  do
  {
    cerr << "Trying to disconnect from server...\n";
    ret = pbs_disconnect (m_pbs_stream);
  } while (ret != 0);
  cerr << "Disconnected...\n";
}  


void
pbs_harness::create_pbs_script (benchmark_info *bench,
				genetic_expr *expr,
				uint32 expr_num)
//////////////////////////////////////////////////////////////
//
//  Creates and runs a PBS script that compiles the bench-
//  mark with the given expression.
//
//////////////////////////////////////////////////////////////
{
  string pbsfname = string("b") + num_string(bench->bid());
  pbsfname += string("_e") + num_string(expr_num);

  ofstream pbsfile(pbsfname.c_str());
  while (!pbsfile) {
    cerr << "Could not create " << pbsfname;
    cerr << " ; Retrying.\n";
    sleep (5);
    pbsfile.open (pbsfname.c_str());
  }

  pbsfile << "#!/bin/sh\n\n";

  string expr_dir = eval_dir() + num_string(bench->bid());
  mkdir(expr_dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
  expr_dir += string("/") + num_string(expr_num);
  mkdir(expr_dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
  string pfun_dir = expr_dir + string("/evaluate/");
  mkdir(pfun_dir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
  
  pbsfile << "export FINCH_EVAL_DIR=" << expr_dir << endl;
  pbsfile << "export FINCH_HOME=" << workdir() << endl;
  
  string pfun_fname = pfun_dir + string("pfun.l");
  pbsfile << "cd " <<  pfun_dir << endl;
  
  ofstream pfun(pfun_fname.c_str());
  expr->print (pfun);
  pfun.close();

  string lock_fname = pfun_dir + ("lock");
  finch_unlink (lock_fname, NULL);
  
  write_commands (pbsfile, bench->get_setup_commands());
  write_commands (pbsfile, bench->get_profile_commands());
  write_commands (pbsfile, bench->get_compile_commands());
  write_commands (pbsfile, bench->get_link_commands());
  write_commands (pbsfile, bench->get_run_commands());
  write_commands (pbsfile, bench->get_check_commands());

  pbsfile << "sync\n";
  pbsfile << "sync\n";
  pbsfile << "sync\n";
  pbsfile << "touch " << lock_fname << endl;
  
  pbsfile.close();
  submit_script (pbsfname);
}


void
pbs_harness::submit_script (const string& name)
{
  struct attropl *pbs_attrs;
  char nname[256];
  char *jobid;

  sync ();
  sync ();

  strcpy (nname, name.c_str());
  pbs_attrs = pbs_attr_push_front (m_common_attrs, ATTR_N, nname);

  do
  {
    cerr << "Trying to submit: (errno=";
    jobid = pbs_submit (m_pbs_stream, pbs_attrs, nname, NULL, NULL);
    cerr << pbs_errno << ")\n";
  } while (jobid == NULL);

  cerr << "Submitted " << name << ": " << jobid << endl;
  free (jobid);
  free (pbs_attrs);
}


struct attropl*
pbs_harness::pbs_attr_push_front (struct attropl* list,
				  char *name,
				  char *value)
{
  struct attropl *tmp;

  tmp = (struct attropl*) malloc (sizeof(struct attropl));
  tmp->name = name;
  tmp->resource = NULL;
  tmp->value = value;
  tmp->next = list;

  return tmp;
}

void
pbs_harness::create_attr_list ()
{
  struct attropl *list;

  list = pbs_attr_push_front (NULL, ATTR_S, "/bin/sh");
  list = pbs_attr_push_front (list, ATTR_p, "20");
  list = pbs_attr_push_front (list, ATTR_e, "/dev/null");
  list = pbs_attr_push_front (list, ATTR_o, "/dev/null");
  list = pbs_attr_push_front (list, ATTR_M, "nobody@cag.lcs.mit.edu");
  list = pbs_attr_push_front (list, ATTR_m, "n");
  
  m_common_attrs = list;
}

void
pbs_harness::write_commands (ofstream& pbsfile,
			     command_list *cmds)
{
  command_list *beg = cmds;

  while (beg != NULL)
  {
    pbsfile << beg->entry << endl;
    beg = beg->next;
  }
}


bool
pbs_harness::gather_results (benchmark_info* bench,
			     genetic_expr* expr,
			     uint32 expr_num)
{
  FILE   *lockfile;
  double  fitness;
  bool    ret;

  string expr_dir = eval_dir() + num_string(bench->bid());
  expr_dir += string("/") + num_string(expr_num);
  string fitness_dir = expr_dir + string("/evaluate/");
  string lockfname = fitness_dir + string("lock");
  string fitnessfname = fitness_dir + string("fitness");

  while (!(lockfile = fopen(lockfname.c_str(), "r")))
    sleep (2);
  fclose(lockfile);

  ifstream fitfile (fitnessfname.c_str());
  fitness = NAN;
  if (!fitfile.fail()) {
    fitfile >> fitness;
    fitfile.close ();
  }

  // Make sure that it works according to the user.
  execute_commands (bench->get_check_commands());

  // Assign the fitness to the expression, thereby
  // memoizing it.
  ret = (fitness == 0.0 || isnan(fitness)); 
  if (ret)
    expr->set_fitness_element (bench->bid(), NAN);
  else 
    expr->set_fitness_element (bench->bid(), fitness);
  
  cout << fitness << "\n";
  
  finch_unlink (fitness_dir, NULL);
  string pbsfname = string("rm -f b") + num_string(bench->bid());
  pbsfname += string("_e") + num_string(expr_num);
  pbsfname += "*";
  system (pbsfname.c_str());

  return ret;
}

