/*************************************************************************
 *
 *   File: simple_harness.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Wed Aug 21 19:58:23 2002
 *
 *   Function: A harness that uses the file system to communicate
 *             priority functions with the compiler.
 *
 ************************************************************************/


#include "simple_harness.h"
#include "exec_utils.h"
#include <unistd.h>

simple_harness::simple_harness (uint32  population_size,
				uint32  generations,
				double  mutate_rate,
				double  mortality,
				char*   initial_pop_fname,
				char*   save_pop_fname,
				char*   finch_home,
				uint32  tournament_size,
				uint32  dss_set_size,
				uint32  maximum_genome_size,
				uint32  maximum_height,
				uint32  elitist,
				uint32  sunset,
				bool    quantize_results,
				bool    classifier,
				benchmark_set* benches) :
  compile_harness (population_size,
		   generations,
		   mutate_rate,
		   mortality,
		   initial_pop_fname,
		   save_pop_fname,
		   finch_home,
		   tournament_size,
		   dss_set_size,
		   maximum_genome_size,
		   maximum_height,
		   elitist,
		   sunset,
		   quantize_results,
		   classifier,
		   benches)
{
}


simple_harness::~simple_harness ()
/////////////////////////////////////////////////////////////
//
//  Deletes dynamically allocated memory.
//
/////////////////////////////////////////////////////////////
{
}


void
simple_harness::clear_evaluate_dir ()
/////////////////////////////////////////////////////////////
//
// Clears old work done in the work directories.
// This simple harness works by compiling benchmarks one
// at a time in the $FINCH_HOME/evaluate/ directory.  
// This procedure makes sure that this directory is empty
// initially.
//
/////////////////////////////////////////////////////////////
{
  if (!finch_clean_dir (eval_dir(), NULL)) {
    cerr << "Could not delete " << eval_dir() << endl;
    exit (-1);
  }
}


void
simple_harness::farm_out_jobs ()
/////////////////////////////////////////////////////////////
//
//  This procedure runs and collects fitness numbers for
//  all the expressions that haven't aleady been memoized.
//  This procedure is also responsible for determining 
//  and marking memoized expressions.
//
/////////////////////////////////////////////////////////////
{
  benchmark_info *bench;
  uint32 num_benches;
  uint32 bn;
  uint32 en;

  cout << endl;

  // Find the subset of benchmarks.
  benches()->sort ();

  // Summarize the DSS state in the transcript.
  benches()->summarize (cout);

  // Figure out the setsize to use.
  num_benches = dss_size ();

  cout << "Using a set size of " << num_benches 
       << endl << endl;

  // For each benchmark in the subset.
  for (bn = 0; bn < num_benches; bn++)
  {
    // Select a benchmark from the subset.
    bench = benches()->bench (bn);

    // For each expression in the population.
    for (en = 0; en < popsize(); en++)
    {
      // Compile and run the benchmark.
      cout << "<Expr=" << en << ", ";
      dispatch (bench, population()[en]);
      write_state (bench, population()[en], en);
    }
  }

  // Compute the overall fitness and the optimizability
  // of each benchmark in the training subset.
  compute_overall_fitness (num_benches);
  cout << endl;
}  


void
simple_harness::dispatch (benchmark_info *bench, 
			  genetic_expr *expr)
/////////////////////////////////////////////////////////////
//  
//  Actually compiles and runs the benchmark with the given
//  expression.  All fitness bookkeeping is stored in the
//  genetic_expr and the benchmark_info.
//
/////////////////////////////////////////////////////////////
{
  double fitness;
  char *curdir;

  // Check to see if the benchmark hasn't
  // already been memoized.
  cout << "Bench=" << bench->bid() << ", ";
  if (!(expr->memoized(bench->bid())))
  {
    // Save away the current directory and switch to 
    // the compile directory.
    curdir = getcwd (NULL, 0);
    chdir (eval_dir().c_str());

    // Run the setup commands.
    execute_commands (bench->get_setup_commands());

    // If it hasn't writen the expression to 
    // $FINCH_HOME/evaluate/pfun.l.
    string expr_fname = eval_dir() + string ("pfun.l");
    string fit_fname = eval_dir() + string ("fitness");
    ofstream expr_file (expr_fname.c_str());
    expr->print (expr_file);
    expr_file.close ();

    // Execute all the commands needed to profile, compile
    // run, and check the benchmark.  It is the user's 
    // responsibility to make sure that the fitness file
    // is written.
    execute_commands (bench->get_profile_commands());
    execute_commands (bench->get_compile_commands());
    execute_commands (bench->get_link_commands());
    execute_commands (bench->get_run_commands());
    
    ifstream fitfile (fit_fname.c_str());
    fitness = NAN;
    if (!fitfile.fail()) {
      fitfile >> fitness;
      fitfile.close ();
    }

    execute_commands (bench->get_check_commands());

    cout << "Fitness=" << fitness << ">\n";

    // Assign the fitness to the expression, thereby
    // memoizing it.
    if ((fitness == 0.0) || isnan(fitness))
      expr->set_fitness_element (bench->bid(), NAN);
    else 
      expr->set_fitness_element (bench->bid(), fitness);

    clear_evaluate_dir ();

    // Switch back to the harness directory.
    chdir (curdir);
    delete[] curdir;
  }
  else 
  {
    cout << "Memoized=" 
	 << expr->get_fitness_element(bench->bid()) << ">\n";
  }
}
